import inspect
import os
from functools import partial
from itertools import (chain,
                       starmap)
from types import FunctionType
from typing import (Iterable,
                    Type)

from . import (functions,
               catalog,
               namespaces,
               strings,
               file_system)
from .types import NamespaceType
from .utils import fix_code

PARAMETER_TEMPLATE = '{parameter}: {annotation}'
DEFINITION_TEMPLATE = 'def test_{function}({parameters}) -> None:\n'
RESULT_NAME = 'result'
FUNCTION_CALL_TEMPLATE = '{result} = {function}({arguments})\n\n'
RETURN_TYPE_CHECK_TEMPLATE = 'isinstance({result}, {return_type})'
NONE_RETURN_TYPE_CHECK_TEMPLATE = '{result} is None'
ASSERTION_TEMPLATE = 'assert {statement}\n'


def from_functions(module_functions: Iterable[FunctionType],
                   *,
                   namespace: NamespaceType,
                   spaces_count: int) -> str:
    module_functions = list(module_functions)
    dependants_paths = functions.dependants_paths(module_functions,
                                                  namespace=namespace)
    dependants_paths = catalog.modules_objects_paths(dependants_paths)
    imports = chain.from_iterable(starmap(catalog.to_imports,
                                          dependants_paths.values()))
    test_case_factory = partial(from_function,
                                spaces_count=spaces_count,
                                namespace=namespace)
    test_cases = map(test_case_factory, module_functions)
    code_blocks = chain(imports, test_cases)
    return fix_code(''.join(code_blocks))


def from_function(function: FunctionType,
                  *,
                  namespace: NamespaceType,
                  spaces_count: int) -> str:
    signature = functions.signature(function)
    parameters = signature.parameters
    parameters_str = strings.join(
        PARAMETER_TEMPLATE.format(
            parameter=parameter.name,
            annotation=parameter.annotation.to_string(namespace))
        for parameter in parameters)
    try:
        parameter, = parameters
    except ValueError:
        arguments_str = strings.join(
            functions.ARGUMENTS_TEMPLATES[parameter.kind].format(
                parameter=parameter.name,
                argument=parameter.name)
            for parameter in parameters)
    else:
        kind = parameter.kind
        if kind == inspect._POSITIONAL_OR_KEYWORD:
            kind = inspect._POSITIONAL_ONLY
        arguments_str = (functions.ARGUMENTS_TEMPLATES[kind]
                         .format(parameter=parameter.name,
                                 argument=parameter.name))
    return_type = signature.return_type
    return_type_bases = return_type.bases

    def to_statement(annotation_base: Type) -> str:
        if annotation_base is None:
            return NONE_RETURN_TYPE_CHECK_TEMPLATE
        else:
            return RETURN_TYPE_CHECK_TEMPLATE

    statements = sorted(set(map(to_statement, return_type_bases)))
    statement = strings.join_with_wrapping(statements,
                                           sep='\nor\n')
    assertion_template = ASSERTION_TEMPLATE.format(statement=statement)

    return_type_str = None
    defined_return_types_bases = list(filter(None, return_type_bases))
    name_seeker = partial(namespaces.search_name,
                          namespace=namespace)
    if defined_return_types_bases:
        if len(defined_return_types_bases) == 1:
            return_type_base, = defined_return_types_bases
            return_type_str = name_seeker(return_type_base)
        else:
            return_types = list(map(name_seeker,
                                    defined_return_types_bases))
            return_types_str = ', '.join(return_types)
            return_type_str = ('({return_types})'
                               .format(return_types=return_types_str))

    tab = ' ' * spaces_count
    template = (DEFINITION_TEMPLATE
                + tab + FUNCTION_CALL_TEMPLATE
                + tab + assertion_template)
    return template.format(function=function.__name__,
                           parameters=parameters_str,
                           result=RESULT_NAME,
                           arguments=arguments_str,
                           return_type=return_type_str)


def normalize_path(path: str,
                   *,
                   source_extension: str = file_system.SOURCE_EXTENSION
                   ) -> str:
    path = file_system.to_relative(path)
    module_full_name = catalog.path_to_module_path(path)
    *sup_modules, module_name = str(module_full_name).split(catalog.SEPARATOR)
    module_file_name = 'test_' + module_name + source_extension
    return os.path.join(*sup_modules, module_file_name)



import os
from itertools import filterfalse
from typing import Iterable

from . import strings
from .utils import is_python_module


def validate_paths(paths: Iterable[str]) -> None:
    non_existent_paths = list(filterfalse(os.path.exists, paths))

    if not non_existent_paths:
        return

    non_existent_paths_str = strings.join(map(strings.wrap_with_quotes,
                                              non_existent_paths))
    err_msg = ('Next paths not found on file system:\n'
               '{paths}.'
               .format(paths=non_existent_paths_str))
    raise FileNotFoundError(err_msg)


def validate_modules_paths(paths: Iterable[str]) -> None:
    invalid_paths = list(filterfalse(is_python_module, paths))

    if invalid_paths:
        err_msg = ('Next paths are not valid Python modules:\n'
                   '{paths}'.format(paths=strings.join(invalid_paths)))
        raise OSError(err_msg)



import collections
import inspect
import operator
from contextlib import suppress
from datetime import (time,
                      timedelta,
                      date,
                      datetime)
from decimal import Decimal
from fractions import Fraction
from functools import partial
from itertools import (chain,
                       starmap)
from typing import (Any,
                    Type,
                    Iterable,
                    Iterator,
                    Dict,
                    Tuple,
                    List)

from hypothesis import strategies

from . import (functions,
               annotator,
               namespaces,
               parameters,
               catalog,
               strings)
from .annotator.detectors import is_generic
from .types import NamespaceType
from .utils import (fix_code,
                    merge_mappings,
                    to_name)


def init_module(modules_parameters: Dict[catalog.ModulePath,
                                         List[inspect.Parameter]]) -> str:
    strategies_paths = (
        (catalog.ContentPath(module=catalog.ModulePath(catalog.SEPARATOR
                                                       + str(module_path)),
                             object=to_strategy_name(parameter),
                             type=catalog.PathType.relative)
         for parameter in module_parameters)
        for module_path, module_parameters in modules_parameters.items()
    )
    imports = chain.from_iterable(starmap(catalog.to_imports,
                                          strategies_paths))
    source = ''.join(imports)
    return fix_code(source)


def from_parameters(module_parameters: Iterable[inspect.Parameter],
                    *,
                    namespace: NamespaceType) -> str:
    module_parameters = list(module_parameters)
    code_blocks = chain(module_imports(module_parameters,
                                       namespace=namespace),
                        module_strategies_definitions(module_parameters,
                                                      namespace=namespace))
    return fix_code(''.join(code_blocks))


def name_to_module_path(full_name: str) -> catalog.ModulePath:
    return (catalog.name_to_module_path(full_name)
            ._replace(type=catalog.PathType.relative))


utilities = {name_to_module_path(strategies.__name__): strategies}


def module_imports(module_parameters: Iterable[inspect.Parameter],
                   *,
                   namespace: NamespaceType) -> Iterator[str]:
    namespace = merge_mappings(namespace, utilities)
    annotations = map(operator.attrgetter('annotation'), module_parameters)
    objects = set(chain.from_iterable(map(dependant_types, annotations)))
    object_path_seeker = partial(namespaces.search_path,
                                 namespace=namespace)
    objects_paths = map(object_path_seeker, objects)
    modules_objects_paths = (catalog.modules_objects_paths(objects_paths)
                             .values())
    yield from chain.from_iterable(starmap(catalog.to_imports,
                                           modules_objects_paths))


def dependant_types(annotation: annotator.Annotation) -> Iterator[Type]:
    bases = annotation.bases
    yield from chain.from_iterable(templates_dependants
                                   .get(base, [strategies.builds, base])
                                   for base in bases)
    initializers_parameters = chain.from_iterable(
            map(parameters.from_type_initializer, bases))
    annotations = map(operator.attrgetter('annotation'),
                      initializers_parameters)
    yield from chain.from_iterable(map(dependant_types,
                                       annotations))
    if isinstance(annotation, annotator.annotations.Generic):
        arguments = annotation.arguments
        yield from chain.from_iterable(map(dependant_types, arguments))


def module_strategies_definitions(
        module_parameters: Iterable[inspect.Parameter],
        *,
        namespace: NamespaceType) -> Iterator[str]:
    namespace = merge_mappings(namespace, utilities)
    strategy_definition_factory = partial(strategy_definition,
                                          namespace=namespace)
    yield from map(strategy_definition_factory, module_parameters)


def strategy_definition(parameter: inspect.Parameter,
                        *,
                        namespace: NamespaceType) -> str:
    strategy_name = to_strategy_name(parameter)
    template = to_template(parameter.annotation)
    value = template.to_string(namespace)
    return strategy_name + ' = ' + value + '\n'


templates = {
    type: functions.FunctionCall(
            strategies.builds,
            functions.Argument(name='target',
                               value=type),
            functions.FunctionCall(strategies.from_regex,
                                   functions.Argument(name='regex',
                                                      value='\A[_a-zA-Z]+\Z')),
            functions.FunctionCall(strategies.tuples),
            functions.FunctionCall(strategies.builds,
                                   functions.Argument(name='target',
                                                      value=dict))),
    object: functions.FunctionCall(strategies.none),
    int: functions.FunctionCall(strategies.integers),
    bool: functions.FunctionCall(strategies.booleans),
    str: functions.FunctionCall(strategies.text),
    float: functions.FunctionCall(strategies.floats),
    Decimal: functions.FunctionCall(strategies.decimals),
    Fraction: functions.FunctionCall(strategies.fractions),
    None: functions.FunctionCall(strategies.none),
    collections.Hashable: functions.FunctionCall(strategies.integers),
    collections.Iterator: functions.FunctionCall(strategies.iterables),
    collections.Iterable: functions.FunctionCall(strategies.iterables),
    collections.Mapping: functions.FunctionCall(strategies.dictionaries),
    collections.Container: functions.FunctionCall(strategies.lists),
    collections.Sequence: functions.FunctionCall(strategies.lists),
    collections.Collection: functions.FunctionCall(strategies.lists),
    collections.MutableSet: functions.FunctionCall(strategies.lists),
    dict: functions.FunctionCall(strategies.dictionaries),
    tuple: functions.FunctionCall(strategies.tuples),
    frozenset: functions.FunctionCall(strategies.frozensets),
    set: functions.FunctionCall(strategies.sets),
    list: functions.FunctionCall(strategies.lists),
    timedelta: functions.FunctionCall(strategies.timedeltas),
    time: functions.FunctionCall(strategies.times),
    date: functions.FunctionCall(strategies.dates),
    datetime: functions.FunctionCall(strategies.datetimes),
}


def combine(*objects: Any,
            module_path: catalog.ModulePath
            ) -> Iterable[Tuple[catalog.ContentPath, Any]]:
    for object_ in objects:
        content_path = catalog.ContentPath(module=module_path,
                                           object=object_.__name__,
                                           type=catalog.PathType.absolute)
        yield content_path, object_


with suppress(ImportError):
    from hypothesis.extra import numpy
    import numpy as np

    dtype = functions.Argument(name='dtype',
                               value=object,
                               kind=inspect._POSITIONAL_OR_KEYWORD)
    shape = functions.Argument(name='shape',
                               value=0,
                               kind=inspect._POSITIONAL_OR_KEYWORD)
    templates[np.ndarray] = functions.FunctionCall(numpy.arrays,
                                                   dtype,
                                                   shape)
    numpy_module_path = name_to_module_path(numpy.__name__)
    utilities.update(dict(combine(numpy.arrays,
                                  module_path=numpy_module_path)))

with suppress(ImportError):
    from hypothesis.extra import pandas
    import pandas as pd

    names_or_number = functions.Argument(name='names_or_number',
                                         value=1,
                                         kind=inspect._POSITIONAL_OR_KEYWORD)
    elements = functions.Argument(name='elements',
                                  value=templates[object],
                                  kind=inspect._POSITIONAL_OR_KEYWORD)
    columns = functions.Argument(name='columns',
                                 value=functions.FunctionCall(pandas.columns,
                                                              names_or_number,
                                                              elements))
    templates[pd.Series] = functions.FunctionCall(pandas.series,
                                                  elements)
    templates[pd.DataFrame] = functions.FunctionCall(pandas.data_frames,
                                                     columns)
    pandas_module_path = name_to_module_path(pandas.__name__)
    utilities.update(dict(combine(pandas.columns,
                                  pandas.series,
                                  pandas.data_frames,
                                  module_path=pandas_module_path)))

templates_dependants = {key: list(functions.walk(value))
                        for key, value in templates.items()}


def to_template(annotation: annotator.Annotation
                ) -> functions.FunctionCall:
    try:
        base, = annotation.bases
    except ValueError as err:
        if isinstance(annotation, (annotator.annotations.Union,
                                   annotator.annotations.Optional)):
            return functions.FunctionCall(strategies.one_of,
                                          *map(to_template,
                                               annotation.arguments))
        raise err
    try:
        template = templates[base]
    except KeyError:
        initializer_parameters = parameters.from_type_initializer(base)
        arguments = [
            functions.Argument(name=parameter.name,
                               value=to_template(parameter.annotation),
                               kind=inspect._POSITIONAL_OR_KEYWORD)
            for parameter in initializer_parameters]
        template = functions.FunctionCall(strategies.builds,
                                          functions.Argument(name='target',
                                                             value=base),
                                          *arguments)
    else:
        if isinstance(annotation, annotator.annotations.Generic):
            template = functions.FunctionCall(template.function,
                                              *map(to_template,
                                                   annotation.arguments))
    return template


def to_strategy_name(parameter: inspect.Parameter) -> str:
    annotation = parameter.annotation
    annotation_bases = annotation.bases
    parameter_full_name = parameter.name
    try:
        annotation_base, = annotation_bases
    except ValueError:
        pass
    else:
        if is_generic(annotation.origin):
            parameter_full_name += '_' + to_name(annotation_base).lower()
    return strings.to_plurals(parameter_full_name,
                              target_case=strings.Case.snake)



import importlib.util
import inspect
from types import ModuleType
from typing import (Any,
                    Dict)

from . import (catalog,
               file_system)
from .catalog import ObjectPathType


def from_module_path(module_path: catalog.ModulePath) -> ModuleType:
    if module_path.type == catalog.PathType.relative:
        sup_module_full_name = str(module_path.module)
        sup_module = importlib.import_module(sup_module_full_name)
        return getattr(sup_module, module_path.object)
    module_full_name = str(module_path)
    return importlib.import_module(module_full_name)


def from_path(path: str) -> ModuleType:
    object_path = catalog.path_to_module_path(file_system.to_relative(path))
    return from_module_path(object_path)


def search(object_path: ObjectPathType,
           *,
           modules: Dict[catalog.ModulePath, ModuleType]) -> Any:
    if isinstance(object_path, catalog.ModulePath):
        return modules[object_path]
    module = modules[object_path.module]
    if object_path.object is None:
        err_msg = ('Invalid content path: '
                   'object should not be "None".')
        raise ValueError(err_msg)
    return getattr(module, object_path.object)


def is_built_in(module: ModuleType) -> bool:
    return not hasattr(module, '__file__')


def is_object_from_module(object_: Any,
                          *,
                          module: ModuleType) -> bool:
    return inspect.getmodule(object_) is module and object_ is not module



from .utils import (ARGUMENTS_TEMPLATES,
                    FunctionCall,
                    Argument,
                    signature,
                    dependants_paths,
                    walk)



import inspect
import operator
from contextlib import suppress
from functools import partial
from itertools import (chain,
                       filterfalse)
from types import FunctionType
from typing import (Any,
                    Iterable,
                    Iterator,
                    NamedTuple,
                    Dict,
                    Tuple,
                    List)

from liable import (annotator,
                    namespaces,
                    catalog,
                    strings)
from liable.catalog import ObjectPathType
from liable.types import NamespaceType

from .detectors import (supports_to_string,
                        is_literal)

ARGUMENTS_TEMPLATES = {
    inspect._POSITIONAL_ONLY: '{argument}',
    inspect._POSITIONAL_OR_KEYWORD: '{parameter}={argument}',
    inspect._VAR_POSITIONAL: '*{argument}',
    inspect._KEYWORD_ONLY: '{parameter}={argument}',
    inspect._VAR_KEYWORD: '**{argument}'}


class Signature(NamedTuple):
    parameters: List[inspect.Parameter]
    return_type: annotator.Annotation


class Argument:
    def __init__(self,
                 *,
                 name: str,
                 value: Any,
                 kind: inspect._ParameterKind = inspect._POSITIONAL_ONLY):
        self.name = name
        self.value = value
        self.kind = kind

    def to_string(self, namespace: NamespaceType) -> str:
        value = self.value
        if supports_to_string(value):
            value_name = value.to_string(namespace)
        elif is_literal(value):
            if isinstance(value, str):
                value_name = '\'' + value + '\''
            else:
                value_name = str(value)
        else:
            value_name = namespaces.search_name(value,
                                                namespace=namespace)
        return (ARGUMENTS_TEMPLATES[self.kind]
                .format(parameter=self.name,
                        argument=value_name))


class FunctionCall:
    def __init__(self,
                 function: FunctionType,
                 *arguments: Argument):
        self.function = function
        self.arguments = arguments

    def to_string(self, namespace: NamespaceType) -> str:
        function_name = namespaces.search_name(self.function,
                                               namespace=namespace)
        arguments_str = strings.join(parameter.to_string(namespace)
                                     for parameter in self.arguments)
        return ('{name}({arguments})'
                .format(name=function_name,
                        arguments=arguments_str))


def walk(object_: Any) -> Iterator[Any]:
    def is_function_call(object_: Any) -> bool:
        return isinstance(object_, FunctionCall)

    def walk_argument(argument: Argument) -> Iterator[Any]:
        value = argument.value
        if is_literal(value):
            return
        elif is_function_call(value):
            yield from walk_function_call(value)
        else:
            yield value

    def walk_function_call(function_call: FunctionCall) -> Iterator[Any]:
        yield function_call.function

        arguments = function_call.arguments
        sub_calls = filter(is_function_call, arguments)
        yield from chain.from_iterable(map(walk_function_call, sub_calls))
        plain_arguments = filterfalse(is_function_call, arguments)
        yield from chain.from_iterable(map(walk_argument, plain_arguments))

    walkers = {Argument: walk_argument,
               FunctionCall: walk_function_call}

    with suppress(KeyError):
        yield from walkers[type(object_)](object_)


def dependants_paths(functions: Iterable[FunctionType],
                     *,
                     namespace: NamespaceType
                     ) -> Iterator[ObjectPathType]:
    dependencies_detector = partial(dependencies,
                                    namespace=namespace)
    signatures_dependants = chain.from_iterable(map(dependencies_detector,
                                                    functions))
    object_path_seeker = partial(namespaces.search_path,
                                 namespace=namespace)
    result = set(chain(map(object_path_seeker,
                           signatures_dependants)))
    yield from filterfalse(catalog.is_built_in, result)


def dependencies(function: FunctionType,
                 *,
                 namespace: NamespaceType) -> Iterator[Any]:
    yield function
    function_signature = signature(function)
    parameters = function_signature.parameters
    annotations = map(operator.attrgetter('annotation'), parameters)
    annotation_walker = partial(annotator.walk,
                                namespace=namespace)
    yield from chain.from_iterable(map(annotation_walker, annotations))
    return_type = function_signature.return_type
    yield from return_type.bases


def signature(function: FunctionType) -> Signature:
    raw_signature = inspect.signature(function)
    parameters = raw_signature.parameters.values()
    parameters = list(map(normalize_annotation, parameters))
    return_type = annotator.normalize(raw_signature.return_annotation)
    return Signature(parameters=parameters,
                     return_type=return_type)


def normalize_annotation(parameter: inspect.Parameter) -> inspect.Parameter:
    parameter = normalize_parameter(parameter)
    annotation = annotator.normalize(parameter.annotation)
    return parameter.replace(annotation=annotation)


def normalize_parameter(parameter: inspect.Parameter) -> inspect.Parameter:
    if parameter.kind == inspect._VAR_POSITIONAL:
        return parameter.replace(annotation=Tuple[parameter.annotation])
    elif parameter.kind == inspect._VAR_KEYWORD:
        return parameter.replace(annotation=Dict[str, parameter.annotation])
    return parameter



import os
import sys
from functools import partial
from importlib._bootstrap_external import SOURCE_SUFFIXES
from typing import (Iterable,
                    Iterator)

from . import strings

SOURCE_EXTENSION = SOURCE_SUFFIXES[0]
INIT_MODULE_NAME = '__init__'
INIT_MODULE_FILE_NAME = INIT_MODULE_NAME + SOURCE_EXTENSION


def find_files(path: str,
               *,
               recursive: bool) -> Iterator[str]:
    if os.path.isdir(path):
        yield from directory_files(path,
                                   recursive=recursive)
    else:
        yield path


def directory_files(path: str,
                    *,
                    recursive: bool) -> Iterator[str]:
    if recursive:
        for root, _, files_names in os.walk(path):
            yield from map(partial(os.path.join, root), files_names)
    else:
        yield from filter(os.path.isfile,
                          map(partial(os.path.join, path), os.listdir(path)))


def make_packages(directory: str,
                  *sub_directories: str) -> None:
    test_cases_directory = os.path.join(directory, *sub_directories)
    os.makedirs(test_cases_directory,
                exist_ok=True)
    for package_path in strings.iterative_join(directory, *sub_directories,
                                               sep=os.sep):
        make_init_module(package_path)


def make_init_module(directory: str,
                     file_name: str = INIT_MODULE_FILE_NAME) -> None:
    path = os.path.join(directory, file_name)
    open(path, mode='a').close()


def to_relative(path: str,
                *,
                system_paths: Iterable[str] = sys.path) -> str:
    try:
        root_path = max((system_path
                         for system_path in system_paths
                         if path.startswith(system_path)),
                        key=len)
    except ValueError as err:
        err_msg = ('Invalid module path: "{path}". '
                   'No root path found in `Python` system paths.'
                   .format(path=path))
        raise ModuleNotFoundError(err_msg) from err
    return os.path.normpath(os.path.relpath(path, root_path))



import collections
from typing import (TypingMeta,
                    GenericMeta,
                    Any,
                    Type)

from .base import Annotation


def is_callable(type_: Type) -> bool:
    return issubclass(type_, collections.Callable)


def is_none_annotation(annotation: Annotation) -> bool:
    return annotation.bases == (None,)


def is_typing(object_: Any) -> bool:
    return (isinstance(type(object_), TypingMeta)
            or
            isinstance(object_, GenericMeta))


def is_generic(object_: Any) -> bool:
    return isinstance(object_, GenericMeta) and object_.__extra__ is not None



import inspect
from functools import partial
from itertools import (chain,
                       filterfalse,
                       starmap)
from typing import Iterable

import pytest

from . import (annotator,
               namespaces,
               catalog,
               strategies)
from .types import NamespaceType
from .utils import (fix_code,
                    to_name)

DEFINITION_TEMPLATE = ('@{pytest}.{fixture}(scope=\'function\')\n'
                       .format(pytest=to_name(pytest),
                               fixture=to_name(pytest.fixture))
                       + 'def {name}() -> {annotation}:\n')
RETURN_TEMPLATE = 'return {strategies}.{strategy}.example()\n'


def from_parameters(parameters: Iterable[inspect.Parameter],
                    *,
                    namespace: NamespaceType,
                    spaces_count: int,
                    tests_module_name: str,
                    strategies_module_name: str) -> str:
    annotations = chain.from_iterable(annotator.walk(parameter.annotation,
                                                     namespace=namespace)
                                      for parameter in parameters)
    object_path_seeker = partial(namespaces.search_path,
                                 namespace=namespace)
    annotations_paths = map(object_path_seeker, annotations)
    additional_objects_paths = [
        catalog.ModulePath(to_name(pytest)),
        catalog.ModulePath(module=catalog.ModulePath(tests_module_name),
                           object=strategies_module_name,
                           type=catalog.PathType.relative)
    ]
    dependant_objects_paths = chain(additional_objects_paths,
                                    annotations_paths)
    dependant_objects_paths = filterfalse(catalog.is_built_in,
                                          dependant_objects_paths)
    modules_dependant_objects_paths = (
        catalog.modules_objects_paths(dependant_objects_paths).values())
    imports = chain.from_iterable(starmap(catalog.to_imports,
                                          modules_dependant_objects_paths))
    fixture_factory = partial(from_parameter,
                              strategies_module_name=strategies_module_name,
                              spaces_count=spaces_count,
                              namespace=namespace)
    fixtures = map(fixture_factory, parameters)
    code_blocks = chain(imports,
                        fixtures)
    return fix_code(''.join(code_blocks))


def from_parameter(parameter: inspect.Parameter,
                   *,
                   namespace: NamespaceType,
                   strategies_module_name: str,
                   spaces_count: int) -> str:
    strategy_name = strategies.to_strategy_name(parameter)
    annotation = parameter.annotation
    tab = ' ' * spaces_count
    return (DEFINITION_TEMPLATE.format(name=parameter.name,
                                       annotation=annotation.to_string(
                                               namespace))
            + tab + RETURN_TEMPLATE.format(strategies=strategies_module_name,
                                           strategy=strategy_name))



import inspect
from typing import (Any,
                    Type,
                    Tuple)


def supports_to_string(object_: Any) -> bool:
    to_string = getattr(object_, 'to_string', None)
    return inspect.ismethod(to_string)


def is_literal(object_: Any,
               *,
               basic_types: Tuple[Type] = (int, bool, float, str)) -> bool:
    return isinstance(object_, basic_types)



import ast
import inspect
import operator
import os
from functools import partial
from itertools import (repeat,
                       starmap,
                       filterfalse)
from types import ModuleType
from typing import (Union,
                    Callable,
                    Iterator,
                    Tuple)

from . import (catalog,
               modules,
               file_system)
from .catalog import ObjectPathType

ImportType = Union[ast.Import, ast.ImportFrom]

ALL_OBJECTS_WILDCARD = '*'

DYNAMIC_MODULES_EXTENSIONS = ('.so', '.dylib', '.dll')


def from_source(source: str,
                *,
                file_name: str = '<unknown>',
                mode: str = 'exec') -> ast.AST:
    return compile(source, file_name, mode, ast.PyCF_ONLY_AST)


def from_module(module: ModuleType) -> ast.AST:
    source = to_source(module)
    return from_source(source,
                       file_name=module.__file__)


def to_source(module: ModuleType,
              *,
              permissible_extensions: Tuple[str] = DYNAMIC_MODULES_EXTENSIONS
              ) -> str:
    module_path = module.__file__
    if os.stat(module_path).st_size == 0:
        return ''
    try:
        return inspect.getsource(module)
    except OSError as err:
        if module_path.endswith(permissible_extensions):
            return ''
        raise err


def to_objects_paths(statement: ImportType,
                     *,
                     all_objects_wildcard: str = ALL_OBJECTS_WILDCARD
                     ) -> Iterator[ObjectPathType]:
    name = operator.attrgetter('name')
    names = list(map(name, statement.names))
    objects_are_relative = isinstance(statement, ast.ImportFrom)
    if objects_are_relative:
        if is_import_relative(statement):
            err_msg = ('Import statement should be absolute, '
                       'relative found.')
            raise ValueError(err_msg)

        objects_names = names[:]

        module_full_name = statement.module
        module_path = catalog.name_to_module_path(module_full_name)
        module = modules.from_module_path(module_path)

        def is_sub_module_name(object_name: str) -> bool:
            content = getattr(module, object_name)
            return inspect.ismodule(content)

        if all_objects_wildcard in objects_names:
            objects_names.remove(all_objects_wildcard)
            try:
                imported_objects_names = module.__all__
            except AttributeError:
                imported_objects_names = vars(module).keys()
            objects_names.extend(imported_objects_names)

        modules_paths = repeat(module_path)
        sub_modules_names = filter(is_sub_module_name, objects_names)
        module_path_factory = partial(catalog.ModulePath,
                                      type=catalog.PathType.relative)
        yield from starmap(module_path_factory,
                           zip(modules_paths, sub_modules_names))

        contents_names = filterfalse(is_sub_module_name, objects_names)
        content_path_factory = partial(catalog.ContentPath,
                                       type=catalog.PathType.relative)
        yield from starmap(content_path_factory,
                           zip(modules_paths, contents_names))
    else:
        # all names are modules names
        yield from map(catalog.name_to_module_path, names)


def import_absolutizer(path: str
                       ) -> Callable[[ast.ImportFrom], ast.ImportFrom]:
    """Returns operator which makes ``ast.ImportFrom`` statements absolute."""

    def to_absolute(statement: ImportType) -> ImportType:
        if isinstance(statement, ast.Import):
            return statement
        module_full_name = to_module_full_name(statement)
        return ast.ImportFrom(module=module_full_name,
                              level=0,
                              names=statement.names,
                              lineno=statement.lineno,
                              col_offset=statement.col_offset)

    directory_path = os.path.normpath(os.path.dirname(path))
    directory_relative_path = file_system.to_relative(directory_path)

    def to_module_full_name(statement: ast.ImportFrom) -> str:
        result = statement.module
        if is_import_relative(statement):
            jumps = (os.pardir + os.sep) * (statement.level - 1)
            path = os.path.join(directory_relative_path, jumps, result or '')
            path = os.path.normpath(path)
            result = str(catalog.path_to_module_path(path))
        return result

    return to_absolute


def is_import_statement(node: ast.AST) -> bool:
    return isinstance(node, (ast.Import, ast.ImportFrom))


def is_import_relative(statement: ast.ImportFrom) -> bool:
    return statement.level > 0



import inspect
from itertools import (filterfalse,
                       chain)
from typing import (TypingMeta,
                    Any,
                    Union,
                    Type,
                    Iterator)

from liable.types import NamespaceType
from liable.utils import to_name

from . import annotations
from .base import Annotation
from .detectors import (is_typing,
                        is_generic,
                        is_callable,
                        is_none_annotation)

NoneType = type(None)

ANNOTATIONS_REPLACEMENTS = {inspect._empty: Any}


def normalize(type_: Type) -> Annotation:
    return to_annotation(ANNOTATIONS_REPLACEMENTS.get(type_,
                                                      type_))


def to_annotation(object_: Any) -> Annotation:
    if not is_typing(object_):
        return annotations.Raw(object_)

    if object_ is Any:
        return annotations.Any()

    try:
        origin = object_.__origin__
    except AttributeError:
        return annotations.Raw(object)

    if origin is Union:
        object_.__args__ = tuple(map(none_type_to_none,
                                     object_.__args__))
        arguments = object_.__args__
        arguments_annotations = list(map(to_annotation, arguments))
        if None in arguments:
            arguments_annotations = list(filterfalse(is_none_annotation,
                                                     arguments_annotations))
            return annotations.Optional(origin=object_,
                                        arguments=arguments_annotations)
        return annotations.Union(origin=object_,
                                 arguments=arguments_annotations)

    if is_callable(object_):
        signature_annotations = object_.__args__
        if not signature_annotations:
            return annotations.PlainAnnotation(object_)
        *parameters_annotations, return_type_annotation = signature_annotations
        return annotations.Callable(
                origin=object_,
                parameters=list(map(to_annotation, parameters_annotations)),
                return_type=to_annotation(return_type_annotation))
    elif is_generic(object_):
        arguments = object_.__args__
        if not arguments:
            return annotations.PlainGeneric(object_)
        arguments_annotations = list(map(to_annotation, arguments))
        return annotations.Generic(origin=object_,
                                   arguments=arguments_annotations)

    return annotations.PlainAnnotation(object_)


def walk(annotation: Annotation,
         *,
         namespace: NamespaceType) -> Iterator[Any]:
    origin = annotation.origin
    if origin in namespace.values():
        yield origin
        return
    yield from walk_plain(annotation)


def walk_plain(annotation: Annotation) -> Iterator[Union[Type, TypingMeta]]:
    if not isinstance(annotation, Annotation):
        err_msg = ('"{cls}" is not an annotation class.'
                   .format(cls=to_name(annotation.__class__)))
        raise TypeError(err_msg)

    if isinstance(annotation, (annotations.Raw,
                               annotations.PlainAnnotation,
                               annotations.Any,
                               annotations.PlainGeneric)):
        yield annotation.origin
        yield from annotation.bases
    elif isinstance(annotation, (annotations.Union, annotations.Optional)):
        yield annotation.origin.__origin__
        yield from chain.from_iterable(map(walk_plain, annotation.arguments))
    elif isinstance(annotation, annotations.Generic):
        yield annotation.origin.__origin__
        yield from chain.from_iterable(map(walk_plain, annotation.arguments))
    elif isinstance(annotation, annotations.Callable):
        yield annotation.origin.__origin__
        yield from chain.from_iterable(map(walk_plain, annotation.parameters))
        yield from walk_plain(annotation.return_type)
    else:
        err_msg = ('Do not know how to walk through '
                   '"{cls}" annotation class '
                   'object "{object}".'
                   .format(cls=to_name(annotation.__class__),
                           object=to_name(annotation)))
        raise TypeError(err_msg)


def none_type_to_none(object_: Any) -> Any:
    if object_ is NoneType:
        return None
    return object_



import inspect
import operator
from collections import defaultdict
from itertools import (chain,
                       filterfalse,
                       product,
                       repeat,
                       starmap)
from types import FunctionType
from typing import (Optional,
                    Type,
                    Iterable,
                    Iterator,
                    Dict,
                    List)

from . import (annotator,
               functions,
               catalog,
               namespaces)
from .types import NamespaceType

COMMONS_MODULE_PATH = catalog.ModulePath('utils')


def to_top(module_path: catalog.ModulePath) -> catalog.ModulePath:
    try:
        sup_module = module_path.module
    except AttributeError:
        return module_path
    else:
        return to_top(sup_module)


def combine(parameters: Iterable[inspect.Parameter],
            *,
            namespace: NamespaceType,
            commons_module_path: str = COMMONS_MODULE_PATH
            ) -> Dict[catalog.ModulePath, List[inspect.Parameter]]:
    result = defaultdict(list)
    for parameter in parameters:
        annotation = parameter.annotation
        try:
            cls, = annotation.bases
        except ValueError:
            module_path = commons_module_path
        else:
            path = namespaces.search_path(cls,
                                          namespace=namespace)
            if catalog.is_built_in(path):
                module_path = commons_module_path
            else:
                module_path = path.module
        result[to_top(module_path)].append(parameter)
    return result


def from_type_initializer(type_: Type) -> Iterator[inspect.Parameter]:
    initializer_signature = functions.signature(type_.__init__)
    # ignoring `self`
    initializer_parameters = initializer_signature.parameters[1:]
    yield from filterfalse(is_variadic, initializer_parameters)


def from_functions(module_functions: Iterable[FunctionType]
                   ) -> Iterator[inspect.Parameter]:
    signatures = list(map(functions.signature, module_functions))
    parameters = chain.from_iterable(map(operator.attrgetter('parameters'),
                                         signatures))
    result = {}
    for parameter in parameters:
        name = parameter.name
        annotation = parameter.annotation
        try:
            previous_parameter = result[name]
        except KeyError:
            result[name] = parameter
            continue
        previous_annotation = previous_parameter.annotation
        if not are_annotations_consistent(annotation, previous_annotation):
            err_msg = ('Invalid parameter: "{parameter}", '
                       'different annotations should agree, '
                       'but found "{previous_annotation}", "{annotation}".'
                       .format(parameter=name,
                               previous_annotation=previous_annotation.origin,
                               annotation=annotation.origin))
            raise ValueError(err_msg)
        if is_annotation_more_specific(annotation, previous_annotation):
            result[name] = parameter
    yield from result.values()


def is_annotation_more_specific(annotation: annotator.Annotation,
                                other_annotation: annotator.Annotation
                                ) -> bool:
    sub_types = set(bases_mro(annotation))
    sub_types -= {object}
    other_sub_types = set(bases_mro(other_annotation))
    other_sub_types -= {object}
    for type_ in sub_types:
        sibling = next(find_siblings(type_,
                                     other_types=other_sub_types))
        other_annotation_is_more_specific = issubclass(sibling, type_)
        if other_annotation_is_more_specific:
            return False
    return True


def are_annotations_consistent(annotation: annotator.Annotation,
                               previous_annotation: annotator.Annotation
                               ) -> bool:
    sub_types = set(bases_mro(annotation))
    sub_types -= {object}
    other_sub_types = set(bases_mro(previous_annotation))
    other_sub_types -= {object}
    return (are_type_systems_related(sub_types, other_sub_types) or
            are_type_systems_related(other_sub_types, sub_types))


def are_type_systems_related(types: Iterable[Type],
                             other_types: Iterable[Type]) -> bool:
    other_types = list(other_types)
    return all(has_sibling(type_,
                           other_types=other_types)
               for type_ in types)


def has_sibling(type_: Type,
                *,
                other_types: Iterable[Type]) -> bool:
    siblings = find_siblings(type_,
                             other_types=other_types)
    sibling = next(siblings, None)
    return sibling is not None


def find_siblings(type_: Type,
                  *,
                  other_types: Iterable[Type]) -> Iterator[Type]:
    other_types = list(other_types)
    types_pairs = product(repeat(type_,
                                 times=len(other_types)),
                          other_types)
    sup_classes = starmap(to_sup_class, types_pairs)
    yield from filter(None, sup_classes)


def to_sup_class(type_: Type, other_type: Type) -> Optional[Type]:
    if issubclass(type_, other_type):
        return other_type
    elif issubclass(other_type, type_):
        return type_
    else:
        return None


def bases_mro(annotation: annotator.Annotation) -> Iterator[Type]:
    yield from chain.from_iterable(map(operator.attrgetter('__mro__'),
                                       annotation.bases))


VARIADIC_PARAMETERS_KINDS = {inspect._VAR_POSITIONAL,
                             inspect._VAR_KEYWORD}


def is_variadic(parameter: inspect.Parameter) -> bool:
    return parameter.kind in VARIADIC_PARAMETERS_KINDS



import builtins
import inspect
from functools import partial
from itertools import chain
from types import (FunctionType,
                   ModuleType)
from typing import (Any,
                    Iterable,
                    Iterator,
                    Tuple)

from . import (catalog,
               modules,
               arboretum)
from .catalog import ObjectPathType
from .types import NamespaceType
from .utils import (to_name,
                    merge_mappings)


def from_module(module: ModuleType) -> NamespaceType:
    return merge_mappings(dependent_objects(module),
                          inner_objects(module))


def built_ins(module: ModuleType = builtins) -> NamespaceType:
    raw_namespace = dict(vars(module))
    raw_namespace['...'] = raw_namespace.pop('Ellipsis')
    return {catalog.guess_type(content)(module=catalog.BUILT_INS_MODULE_PATH,
                                        object=name,
                                        type=catalog.PathType.inner): content
            for name, content in raw_namespace.items()}


def dependent_objects(module: ModuleType) -> NamespaceType:
    if modules.is_built_in(module):
        return {}
    objects_paths = dependent_objects_paths(module)
    return dict(load_dependent_objects(objects_paths))


def dependent_objects_paths(module: ModuleType) -> Iterator[ObjectPathType]:
    tree = arboretum.from_module(module)
    imports = filter(arboretum.is_import_statement, tree.body)
    module_path = module.__file__
    relative_import_to_absolute = arboretum.import_absolutizer(module_path)
    imports = map(relative_import_to_absolute, imports)
    yield from chain.from_iterable(map(arboretum.to_objects_paths, imports))


def load_dependent_objects(objects_paths: Iterable[ObjectPathType]
                           ) -> Iterator[Tuple[ObjectPathType, Any]]:
    objects_paths = list(objects_paths)
    dependencies_paths = map(catalog.to_module_path, objects_paths)
    dependencies_paths = set(dependencies_paths)
    dependencies = dict(zip(dependencies_paths,
                            map(modules.from_module_path, dependencies_paths)))
    objects_seeker = partial(modules.search,
                             modules=dependencies)
    yield from zip(objects_paths,
                   map(objects_seeker, objects_paths))


def inner_objects(module: ModuleType) -> NamespaceType:
    module_full_name = module.__name__
    module_path = catalog.name_to_module_path(module_full_name)

    return {catalog.guess_type(content)(module=module_path,
                                        object=object_name,
                                        type=catalog.PathType.inner): content
            for object_name, content in vars(module).items()
            if modules.is_object_from_module(content,
                                             module=module)}


def search_name(object_: Any,
                *,
                namespace: NamespaceType) -> str:
    path = search_path(object_,
                       namespace=namespace)
    if path.type != catalog.PathType.absolute:
        return path.object
    module_path = path.module
    if module_path.type == catalog.PathType.relative:
        return module_path.object + catalog.SEPARATOR + path.object
    return str(path)


def search_path(object_: Any,
                *,
                namespace: NamespaceType) -> ObjectPathType:
    object_paths = search_paths(object_,
                                namespace=namespace)
    try:
        return next(object_paths)
    except StopIteration as err:
        raise LookupError('Object "{object}" not found in namespace.'
                          .format(object=to_name(object_))) from err


def search_paths(object_: Any,
                 *,
                 namespace: NamespaceType) -> Iterator[ObjectPathType]:
    is_relative = is_object_relative(object_,
                                     namespace=namespace)
    if is_relative:
        yield from search_relative_objects(object_,
                                           namespace=namespace)
    else:
        yield from search_absolute_paths(object_,
                                         namespace=namespace)


def search_absolute_paths(object_: Any,
                          *,
                          namespace: NamespaceType
                          ) -> Iterator[ObjectPathType]:
    sub_modules_by_paths = dict(namespace_modules(namespace))
    sub_namespaces = list(map(from_module, sub_modules_by_paths.values()))
    yield from chain.from_iterable(
            search_absolute_objects(object_,
                                    namespace=namespace,
                                    module_path=module_path)
            for module_path, namespace in zip(sub_modules_by_paths.keys(),
                                              sub_namespaces))
    yield from chain.from_iterable(search_absolute_paths(object_,
                                                         namespace=namespace)
                                   for namespace in sub_namespaces)


def is_object_relative(object_: Any,
                       *,
                       namespace: NamespaceType) -> bool:
    """
    Object called **relative** in namespace
    if it is defined in original module
    or imported using ``import from`` statement.

    Object called **absolute** in namespace
    if it is defined in module
    which imported in original one using ``import`` statement.
    """
    return object_ in namespace.values()


def search_relative_objects(object_: Any,
                            *,
                            namespace: NamespaceType) -> ObjectPathType:
    for path, content in namespace.items():
        if content is object_:
            yield path


def search_absolute_objects(object_: Any,
                            *,
                            namespace: NamespaceType,
                            module_path: catalog.ModulePath
                            ) -> ObjectPathType:
    for path, content in namespace.items():
        if content is object_:
            yield catalog.guess_type(content)(module=module_path,
                                              object=path.object,
                                              type=catalog.PathType.absolute)


def namespace_modules(namespace: NamespaceType
                      ) -> Iterator[Tuple[catalog.ModulePath, ModuleType]]:
    for path, content in namespace.items():
        if inspect.ismodule(content):
            yield path, content


def functions_by_path_type(namespace: NamespaceType,
                           *,
                           path_type: catalog.PathType
                           ) -> Iterator[FunctionType]:
    objects_by_path_type = (content
                            for path, content in namespace.items()
                            if path.type == path_type)
    yield from filter(inspect.isfunction, objects_by_path_type)


inner_functions = partial(functions_by_path_type,
                          path_type=catalog.PathType.inner)



import builtins
import enum
import inspect
import operator
import os
from collections import defaultdict
from itertools import filterfalse
from typing import (Any,
                    Union,
                    Optional,
                    Type,
                    Iterable,
                    Iterator,
                    NamedTuple,
                    Dict,
                    Set)

from . import (file_system,
               strings)

SEPARATOR = '.'


class PathType(enum.Enum):
    inner = 'inner'
    absolute = 'absolute'
    relative = 'relative'


IMPORTS_TEMPLATES = {PathType.absolute: 'import {module}\n',
                     PathType.relative: 'from {module} import {objects}\n'}


class ModulePath(NamedTuple):
    module: Union[str, 'ModulePath']
    object: Optional[str] = None
    type: PathType = PathType.absolute

    def __str__(self):
        module_full_name = str(self.module)
        if self.object is None:
            return module_full_name
        else:
            return module_full_name + SEPARATOR + self.object


class ContentPath(NamedTuple):
    module: ModulePath
    object: str
    type: PathType

    def __str__(self):
        module_full_name = str(self.module)
        if is_built_in(self):
            return self.object
        else:
            return module_full_name + SEPARATOR + self.object


BUILT_INS_MODULE_PATH = ModulePath(builtins.__name__)

ObjectPathType = Union[ModulePath, ContentPath]


def is_absolute(object_path: ObjectPathType) -> bool:
    return object_path.type == PathType.absolute


def is_built_in(object_path: ObjectPathType) -> bool:
    return object_path.module == BUILT_INS_MODULE_PATH


def path_to_module_path(path: str) -> ModulePath:
    package_name, file_name = os.path.split(path)
    module_name = to_module_name(file_name)
    if not package_name:
        return ModulePath(module_name)
    if module_name is None:
        return path_to_module_path(package_name)
    return ModulePath(module=path_to_module_path(package_name),
                      object=module_name,
                      type=PathType.absolute)


def name_to_module_path(full_name: str) -> ModulePath:
    try:
        package_name, full_name = full_name.rsplit(SEPARATOR, 1)
    except ValueError:
        return ModulePath(full_name)
    else:
        return ModulePath(module=name_to_module_path(package_name),
                          object=full_name,
                          type=PathType.absolute)


def to_module_name(file_name: str) -> Optional[str]:
    module_name = inspect.getmodulename(file_name)
    if module_name is None:
        return file_name
    if module_name == file_system.INIT_MODULE_NAME:
        return None
    return module_name


def to_imports(*module_objects_paths: ObjectPathType) -> Iterator[str]:
    modules_paths = set(map(operator.attrgetter('module'),
                            module_objects_paths))
    try:
        module_path, = modules_paths
    except ValueError as err:
        if modules_paths:
            modules_names_str = strings.join('"' + name + '"'
                                             for name in modules_paths)
            err_msg = ('Found modules paths for different modules: '
                       '{modules}.'
                       .format(modules=modules_names_str))
        else:
            err_msg = 'No modules paths found.'
        raise ValueError(err_msg) from err

    if is_built_in(module_objects_paths[0]):
        return

    non_absolute_paths = list(filterfalse(is_absolute, module_objects_paths))
    if non_absolute_paths:
        objects_names = sorted(map(operator.attrgetter('object'),
                                   non_absolute_paths))
        objects_names_str = strings.join_with_wrapping(objects_names)
        yield (IMPORTS_TEMPLATES[PathType.relative]
               .format(module=module_path,
                       objects=objects_names_str))
    absolute_path = next(filter(is_absolute, module_objects_paths), None)
    if absolute_path is not None:
        yield (IMPORTS_TEMPLATES[PathType.absolute]
               .format(module=module_path))


def modules_objects_paths(objects_paths: Iterable[ObjectPathType]
                          ) -> Dict[str, Set[ObjectPathType]]:
    result = defaultdict(list)
    for object_path in objects_paths:
        module_path = to_module_path(object_path)
        if module_path.type == PathType.relative:
            result[module_path.module].append(module_path)
        else:
            result[module_path].append(object_path)
    return dict(zip(result.keys(), map(set, result.values())))


def to_module_path(object_path: ObjectPathType) -> ModulePath:
    if isinstance(object_path, ModulePath):
        return object_path
    return object_path.module


def guess_type(object_: Any) -> Type[ObjectPathType]:
    if inspect.ismodule(object_):
        return ModulePath
    return ContentPath



import operator
import os
from functools import partial
from itertools import chain
from typing import (Any,
                    Hashable,
                    Mapping)

import autopep8

from . import arboretum


def to_name(object_: Any) -> str:
    try:
        return object_.__qualname__
    except AttributeError:
        try:
            return object_.__name__
        except AttributeError:
            return str(object_)


def is_python_module(path: str) -> bool:
    if not os.path.isfile(path):
        return False
    with open(path) as source_file:
        try:
            source = source_file.read()
        except UnicodeDecodeError:
            return False
    try:
        arboretum.from_source(source,
                              file_name=path)
    except SyntaxError:
        return False
    else:
        return True


def merge_mappings(*mappings: Mapping[Hashable, Any]) -> dict:
    items = chain.from_iterable(map(operator.methodcaller('items'), mappings))
    return dict(items)


fix_code = partial(autopep8.fix_code,
                   options={'aggressive': True,
                            'max_line_length': 79})



import collections
from itertools import chain
from typing import (TypingMeta,
                    Any as AnyType,
                    Union as UnionType,
                    Optional as OptionalType,
                    Type,
                    Tuple,
                    Collection)

from liable import (namespaces,
                    strings)
from liable.types import NamespaceType
from .base import Annotation

ELLIPSIS_STRING = '...'


class Raw(Annotation):
    def __init__(self,
                 origin: Type):
        super().__init__(origin)

    def to_string(self, namespace: NamespaceType) -> str:
        return namespaces.search_name(self.origin,
                                      namespace=namespace)

    @property
    def bases(self) -> Tuple[Type]:
        return self.origin,


class PlainAnnotation(Annotation):
    def __init__(self,
                 origin: TypingMeta):
        super().__init__(origin)

    def to_string(self, namespace: NamespaceType) -> str:
        return namespaces.search_name(self.origin,
                                      namespace=namespace)

    @property
    def bases(self) -> Tuple[Type]:
        return self.origin.__origin__,


class Any(Annotation):
    def __init__(self):
        super().__init__(AnyType)

    def to_string(self, namespace: NamespaceType) -> str:
        return namespaces.search_name(self.origin,
                                      namespace=namespace)

    @property
    def bases(self) -> Tuple[object]:
        return object,


class Union(Annotation):
    def __init__(self,
                 origin: TypingMeta,
                 arguments: Collection[Annotation]):
        super().__init__(origin)
        self.arguments = arguments

    def to_string(self, namespace: NamespaceType) -> str:
        origin = self.origin

        if namespaces.is_object_relative(origin,
                                         namespace=namespace):
            return namespaces.search_name(origin,
                                          namespace=namespace)

        base = origin.__origin__
        base_name = namespaces.search_name(base,
                                           namespace=namespace)
        arguments_names = [argument.to_string(namespace)
                           for argument in self.arguments]
        arguments_name = strings.join(arguments_names,
                                      sep=', ')
        return ('{base}[{arguments}]'
                .format(base=base_name,
                        arguments=arguments_name))

    @property
    def bases(self) -> Tuple[Type, ...]:
        return tuple(chain.from_iterable(annotation.bases
                                         for annotation in self.arguments))


class Optional(Annotation):
    def __init__(self,
                 origin: TypingMeta,
                 arguments: Collection[Annotation]):
        super().__init__(origin)
        self.arguments = arguments

    def to_string(self, namespace: NamespaceType) -> str:
        origin = self.origin

        if namespaces.is_object_relative(origin,
                                         namespace=namespace):
            return namespaces.search_name(origin,
                                          namespace=namespace)

        arguments = self.arguments
        # since "Flat is better than nested."
        #   Union[...types..., None]
        # looks better than
        #   Optional[Union[...types...]]
        origin_cls = OptionalType if len(arguments) == 1 else UnionType
        base_name = namespaces.search_name(origin_cls,
                                           namespace=namespace)
        arguments_names = [annotation.to_string(namespace)
                           for annotation in arguments]
        if len(arguments) > 1:
            arguments_names.append(str(None))
        arguments_name = strings.join(arguments_names,
                                      sep=', ')
        return '{base}[{arguments}]'.format(base=base_name,
                                            arguments=arguments_name)

    @property
    def bases(self) -> Tuple[Type, ...]:
        result = tuple(chain.from_iterable(annotation.bases
                                           for annotation in self.arguments))
        return result + (None,)


class Callable(Annotation):
    def __init__(self,
                 origin: TypingMeta,
                 parameters: Collection[Annotation],
                 return_type: Annotation):
        super().__init__(origin)
        self.parameters = parameters
        self.return_type = return_type

    def to_string(self, namespace: NamespaceType) -> str:
        origin = self.origin

        if namespaces.is_object_relative(origin,
                                         namespace=namespace):
            return namespaces.search_name(origin,
                                          namespace=namespace)

        base = origin.__origin__
        base_name = namespaces.search_name(base,
                                           namespace=namespace)
        parameters_names = [parameter.to_string(namespace)
                            for parameter in self.parameters]
        parameters_name = strings.join(parameters_names,
                                       sep=', ')
        return_type_name = self.return_type.to_string(namespace)
        if parameters_name == ELLIPSIS_STRING:
            return ('{origin}[{parameters}, {return_type}]'
                    .format(origin=base_name,
                            parameters=parameters_name,
                            return_type=return_type_name))
        return ('{origin}[[{parameters}], {return_type}]'
                .format(origin=base_name,
                        parameters=parameters_name,
                        return_type=return_type_name))

    @property
    def bases(self) -> Tuple[Type[collections.Callable]]:
        return self.origin.__extra__,


class PlainGeneric(Annotation):
    def __init__(self,
                 origin: TypingMeta):
        super().__init__(origin)

    def to_string(self, namespace: NamespaceType) -> str:
        return namespaces.search_name(self.origin,
                                      namespace=namespace)

    @property
    def bases(self) -> Tuple[Type]:
        return self.origin.__extra__,


class Generic(Annotation):
    def __init__(self,
                 origin: TypingMeta,
                 arguments: Collection[Annotation]):
        super().__init__(origin)
        self.arguments = arguments

    def to_string(self, namespace: NamespaceType) -> str:
        origin = self.origin

        if namespaces.is_object_relative(origin,
                                         namespace=namespace):
            return namespaces.search_name(origin,
                                          namespace=namespace)

        base = origin.__origin__
        base_name = namespaces.search_name(base,
                                           namespace=namespace)
        arguments_names = [argument.to_string(namespace)
                           for argument in self.arguments]
        arguments_name = strings.join(arguments_names,
                                      sep=', ')
        return ('{base}[{elements}]'
                .format(base=base_name,
                        elements=arguments_name))

    @property
    def bases(self) -> Tuple[Type[collections.Iterable]]:
        return self.origin.__extra__,



import os

from setuptools import (setup,
                        find_packages)

project_base_url = 'https://github.com/lycantropos/liable/'

setup_requires = [
    'pytest-runner>=3.0'
]
install_requires = [
    'click>=6.7',
    'pathspec>=0.5.5',
    'autopep8>=1.3.3',
    'inflect>=0.2.5',
    'nltk>=3.2.5',
    'hypothesis>=3.38.5',
    'pytest>=3.3.0',
]
tests_require = [
    'pydevd>=1.1.1',  # debugging
    'pytest>=3.3.0',
    'pytest-cov>=2.5.1',
    'hypothesis>=3.38.5',
]

setup(name='liable',
      packages=find_packages(exclude=('tests',)),
      scripts=[os.path.join('scripts', 'liable')],
      version='0.0.1',
      description='Auto-tests generator.',
      long_description=open('README.md').read(),
      author='Azat Ibrakov',
      author_email='azatibrakov@gmail.com',
      url=project_base_url,
      download_url=project_base_url + 'archive/master.zip',
      setup_requires=setup_requires,
      install_requires=install_requires,
      tests_require=tests_require)



from abc import (ABC,
                 abstractmethod)
from typing import (TypingMeta,
                    Union,
                    Type,
                    Tuple)

from liable.types import NamespaceType


class Annotation(ABC):
    def __init__(self, origin: Union[Type, TypingMeta]):
        self.origin = origin

    @abstractmethod
    def to_string(self, namespace: NamespaceType) -> str:
        raise NotImplemented

    @property
    @abstractmethod
    def bases(self) -> Tuple[Type, ...]:
        raise NotImplemented



import enum
from itertools import (chain,
                       zip_longest)
from typing import (Any,
                    Iterable,
                    Iterator,
                    Sequence)

import inflect
from nltk.corpus import wordnet

STRINGS_SEPARATOR = ',\n'


class Case(enum.Enum):
    upper = 'UPPER_CASE'
    camel = 'camelCase'
    snake = 'snake_case'


def join(strings: Iterable[str],
         *,
         sep: str = STRINGS_SEPARATOR) -> str:
    return sep.join(strings)


def join_with_wrapping(strings: Sequence[str],
                       *,
                       sep: str = STRINGS_SEPARATOR) -> str:
    if len(strings) == 1:
        string, = strings
        return string
    return '(' + sep.join(strings) + ')'


def wrap_with_quotes(string: str) -> Any:
    quote_character = '"'
    return quote_character + string + quote_character


def to_plurals(string: str,
               *,
               target_case: Case = None) -> str:
    if target_case is None:
        target_case = guess_case(string)
    words = split_words(string)
    pluralized_words = map(to_plural, words)
    return to_case_converters[target_case](pluralized_words)


def guess_case(string: str) -> Case:
    if is_snake_case(string):
        return Case.snake
    elif is_upper_case(string):
        return Case.upper
    elif is_camel_case(string):
        return Case.camel
    else:
        raise ValueError('Cannot guess case for string: "{string}".'
                         .format(string=string))


def split_words(string: str) -> Iterator[str]:
    alphanumeric_string = ''.join(character
                                  if character.isalnum()
                                  else ' '
                                  for character in string)
    result = filter(None, alphanumeric_string.split())
    yield from chain.from_iterable(map(split_camel_case, result))


def split_camel_case(string: str) -> Iterator[str]:
    if not string:
        yield string
        return

    word = string[0]
    for character, next_character in zip_longest(string, string[1:],
                                                 fillvalue=''):
        if character.islower() and next_character.isupper():
            yield word.lower()
            word = ''
        if character.isupper() and next_character.islower():
            # for names like "HTTPResponse"
            # which consists of words "HTTP", "Response"
            previous_word = word[:-1]
            if previous_word:
                yield previous_word.lower()
            word = character + next_character
        else:
            word += next_character
    yield word.lower()


def is_camel_case(string: str) -> bool:
    return (not is_snake_case(string)
            and not is_upper_case(string)
            and all(character.isalnum() for character in string))


def is_snake_case(string: str) -> bool:
    return all(not character.isalpha() or character.islower()
               for character in string)


def is_upper_case(string: str) -> bool:
    return all(not character.isalpha() or character.isupper()
               for character in string)


def to_plural(word: str,
              *,
              engine: inflect.engine = inflect.engine()) -> str:
    is_plural = engine.singular_noun
    if not is_noun(word) or is_plural(word):
        return word
    return engine.plural(word)


def to_upper(words: Iterable[str]) -> str:
    return '_'.join(map(str.upper, words))


def to_camel(words: Iterable[str]) -> str:
    words = iter(words)
    result = next(words).lower()
    for word in words:
        if word.isupper():
            result += word
        else:
            result += word.title()
    return result


def to_snake(words: Iterable[str]) -> str:
    return '_'.join(map(str.lower, words))


to_case_converters = {
    Case.upper: to_upper,
    Case.camel: to_camel,
    Case.snake: to_snake
}


def is_noun(word: str) -> bool:
    synsets = wordnet.synsets(word)
    return any(synset.pos() == wordnet.NOUN
               for synset in synsets)


def iterative_join(*strings: str,
                   sep: str = '') -> Iterator[str]:
    if not strings:
        return
    strings = iter(strings)
    result = next(strings)
    yield result
    for string in strings:
        result = result + sep + string
        yield result
